
#include "fileio.h"

#define FILENAME_LENGTH 16

typedef struct {
	char filename[FILENAME_LENGTH];
	uint32_t size;
	uint32_t offset;
} GCC_PACK fileio_slot_t;

static int file_slot_count;
static const uint8_t *file_slot_data;
static const fileio_slot_t *file_slot_table;

void fio_init(const char *data_path) {
	static const uint8_t *search_addr_start = (const uint8_t *)0x08030000;
	static const uint8_t *search_addr_limit = (const uint8_t *)0x0A000000;
	const uint8_t *p;
	file_slot_count = 0;
	for (p = search_addr_start; p < search_addr_limit; p += 256) {
		if (p[0] == 'B' && p[1] == 'O' && p[2] == 'U' && p[3] == 'M' && READ_BE_UINT32(p + 4) == 1) {
			/* hdr+8 (BE32) : timestamp */
			file_slot_count = READ_BE_UINT32(p + 12);
			file_slot_table = (const fileio_slot_t *)(p + 16);
			file_slot_data = (const uint8_t *)(p + 16 + file_slot_count * sizeof(fileio_slot_t));
			break;
		}
	}
	debug(DBG_FILEIO, "fio_init() file_slot_count=%d", file_slot_count);
	if (file_slot_count == 0) {
		error("No files found");
	}
}

static int file_slot_compare(const void *a, const void *b) {
	const char *key = (const char *)a;
	const fileio_slot_t *file_slot = (const fileio_slot_t *)b;
	return strcmp(key, file_slot->filename);
}

int fio_open(const char *filename, fio_open_mode_e mode, int error_flag) {
	int slot = -1;
	if (mode == FIO_READ) {
		fileio_slot_t *file_slot = bsearch(filename, file_slot_table, file_slot_count, sizeof(fileio_slot_t), file_slot_compare);
		if (!file_slot) {
			if (error_flag) {
				error("Unable to open file '%s'", filename);
			}
			warning("Unable to open file '%s'", filename);
		} else {
			slot = file_slot - file_slot_table;
		}
	} else {
		/* write to SRAM ? */
		error("Unhandled mode %d", mode);
	}
	return slot;
}

void fio_close(int slot) {
	/* nothing to do */
}

int fio_size(int slot) {
	int size = 0;
	if (slot >= 0 && slot < file_slot_count) {
		size = READ_BE_UINT32(&file_slot_table[slot].size);
	}
	return size;
}

const uint8_t *fio_fetch(int slot) {
	const uint8_t *data = 0;
	if (slot >= 0 && slot < file_slot_count) {
		data = file_slot_data + READ_BE_UINT32(&file_slot_table[slot].offset);
	}
	return data;
}

int fio_read(int slot, void *data, int len) {
	if (slot >= 0 && slot < file_slot_count) {
		const uint32_t offs = READ_BE_UINT32(&file_slot_table[slot].offset);
		const uint32_t size = READ_BE_UINT32(&file_slot_table[slot].size);
		if (len > size) {
			len = size;
		}
		memcpy(data, file_slot_data + offs, len);
		return len;
	}
	return 0;
}

int fio_write(int slot, void *data, int len) {
	error("Unhandled write");
	return 0;
}

int fio_exists(const char *filename) {
	return bsearch(filename, file_slot_table, file_slot_count, sizeof(fileio_slot_t), file_slot_compare) != 0;
}
