
#include <sys/param.h>
#include "fileio.h"
#include "sqx_decoder.h"

#define MAX_FILEIO_SLOTS 30

typedef struct {
	int used;
	FILE *fp;
	int size;
	uint8_t *data;
	fio_open_mode_e mode;
	int sqx_compressed;
} fileio_slot_t;

static const char *data_directory;
static fileio_slot_t fileio_slots_table[MAX_FILEIO_SLOTS];

static int find_free_slot() {
	int i, slot = -1;
	for (i = 0; i < MAX_FILEIO_SLOTS; ++i) {
		if (!fileio_slots_table[i].used) {
			slot = i;
			break;
		}
	}
	return slot;
}

void fio_init(const char *data_path) {
	data_directory = data_path;
	memset(fileio_slots_table, 0, sizeof(fileio_slots_table));
}

static int is_sqx_compressed(const char *filename) {
	const char *ext = strrchr(filename, '.');
	if (ext && strcmp(ext + 1, "SQX") == 0) {
		return 1;
	}
	return 0;
}

static FILE *fio_open_helper(const char *filename, fio_open_mode_e mode) {
	static const char *modes_table[] = { "rb", "wb" };
	const char *path = "";
	if (is_sqx_compressed(filename)) {
		if (filename[0] == 'I' || filename[0] == 'S') {
			path = "/SPRITE";
		} else {
			path = "/LEVEL";
		}
	}
	char buf[MAXPATHLEN];
	snprintf(buf, sizeof(buf), "%s%s/%s", data_directory, path, filename);
	return fopen(buf, modes_table[mode]);
}

static void fio_set_sqx_compressed(fileio_slot_t *file_slot, const char *filename) {
	file_slot->sqx_compressed = is_sqx_compressed(filename);
}

int fio_open(const char *filename, fio_open_mode_e mode, int error_flag) {
	int slot = find_free_slot();
	if (slot < 0) {
		error("Unable to find free slot for '%s'", filename);
	} else {
		debug(DBG_FILEIO, "fio_open() slot %d name '%s'", slot, filename);
		fileio_slot_t *file_slot = &fileio_slots_table[slot];
		memset(file_slot, 0, sizeof(fileio_slot_t));
		file_slot->fp = fio_open_helper(filename, mode);
		if (!file_slot->fp) {
			if (error_flag) {
				error("Unable to open file '%s'", filename);
			}
			warning("Unable to open file '%s'", filename);
			slot = -1;
		} else {
			file_slot->mode = mode;
			fseek(file_slot->fp, 0, SEEK_END);
			file_slot->size = ftell(file_slot->fp);
			fseek(file_slot->fp, 0, SEEK_SET);
			file_slot->used = 1;
			fio_set_sqx_compressed(file_slot, filename);
		}
	}
	return slot;
}

void fio_close(int slot) {
	if (slot >= 0) {
		fileio_slot_t *file_slot = &fileio_slots_table[slot];
		assert(file_slot->used);
		free(file_slot->data);
		fclose(file_slot->fp);
		memset(file_slot, 0, sizeof(fileio_slot_t));
	}
}

int fio_size(int slot) {
	int size = 0;
	if (slot >= 0) {
		fileio_slot_t *file_slot = &fileio_slots_table[slot];
		assert(file_slot->used);
		if (file_slot->sqx_compressed && !file_slot->data) {
			fio_fetch(slot);
		}
		size = file_slot->size;
	}
	return size;
}

const uint8_t *fio_fetch(int slot) {
	uint8_t *data = 0;
	if (slot >= 0) {
		fileio_slot_t *file_slot = &fileio_slots_table[slot];
		assert(file_slot->used && file_slot->mode == FIO_READ);
		if (!file_slot->data) {
			file_slot->data = (uint8_t *)malloc(file_slot->size);
			if (!file_slot->data) {
				error("Unable to allocate %d bytes", file_slot->size);
			} else {
				int sz, pos = ftell(file_slot->fp);
				fseek(file_slot->fp, 0, SEEK_SET);
				sz = fread(file_slot->data, 1, file_slot->size, file_slot->fp);
				fseek(file_slot->fp, pos, SEEK_SET);
				if (sz != file_slot->size) {
					error("Failed to read file slot %d (count %d, read %d)", slot, file_slot->size, sz);
					free(file_slot->data);
					file_slot->data = 0;
				} else if (file_slot->sqx_compressed) {
					const int decoded_size = file_slot->size - 6 + (int16_t)READ_LE_UINT16(file_slot->data);
					debug(DBG_FILEIO, "fio_fetch() slot %d decoded_size %d", slot, decoded_size);
					uint8_t *dst = (uint8_t *)malloc(decoded_size);
					if (!dst) {
						error("Unable to allocate %d bytes", decoded_size);
						free(file_slot->data);
						file_slot->data = 0;
					} else {
						const int size = sqx_decode(file_slot->data + 2, dst);
						if (size != decoded_size) {
							error("Failed to decode file_slot %d", slot);
							free(dst);
							dst = 0;
						}
						free(file_slot->data);
						file_slot->data = dst;
						file_slot->size = size;
					}
				}
			}
		}
		data = file_slot->data;
	}
	return data;
}

int fio_read(int slot, void *data, int len) {
	int sz = 0;
	if (slot >= 0) {
		fileio_slot_t *file_slot = &fileio_slots_table[slot];
		assert(file_slot->used && file_slot->mode == FIO_READ);
		if (file_slot->sqx_compressed) {
			assert(len == file_slot->size);
			memcpy(data, file_slot->data, len);
		} else {
			sz = fread(data, 1, len, file_slot->fp);
			if (sz != len) {
				error("Failed to read file slot %d (count %d, read %d)", slot, len, sz);
			}
		}
	}
	return sz;
}

int fio_write(int slot, void *data, int len) {
	int sz = 0;
	if (slot >= 0) {
		fileio_slot_t *file_slot = &fileio_slots_table[slot];
		assert(file_slot->used && file_slot->mode == FIO_WRITE);
		sz = fwrite(data, 1, len, file_slot->fp);
		if (sz != len) {
			error("Failed to write file slot %d (count %d, written %d)", slot, len, sz);
		}
	}
	return sz;
}

int fio_exists(const char *filename) {
	FILE *fp = fio_open_helper(filename, FIO_READ);
	if (fp) {
		fclose(fp);
		return 1;
	}
	return 0;
}
