
#include "fileio.h"
#include "game.h"
#include "globals.h"
#include "resource.h"
#include "systemstub.h"

static const int DEFAULT_SCALE = 2;

struct resource_t g_res;

int load_sprite(const char *name, int level, const uint16_t *pal);
int load_static_sprite(const uint8_t *data, const uint16_t *pal);
void load_tile(const char *name, int level, const uint16_t *pal);
void load_fond(int num, int level, const uint16_t *pal);

void res_detect() {
	if (IS_GBA) {
		g_res.gfx_ext = "TIL";
		g_res.dat_ext = "___";
		g_res.gfx_scale_w = 1;
		g_res.gfx_scale_h = 1;
	} else {
		g_res.converted_data = fio_exists("S00.TXT");
		if (g_res.converted_data) {
			g_res.gfx_ext = "PNG";
			g_res.dat_ext = "DAT";
			g_res.gfx_scale_w = DEFAULT_SCALE;
			g_res.gfx_scale_h = DEFAULT_SCALE;
		} else {
			g_res.gfx_ext = "SQX";
			g_res.dat_ext = "SQX";
			g_res.gfx_scale_w = 1;
			g_res.gfx_scale_h = 1;
		}
	}
	g_res.img = -1;
}

const uint8_t *res_get_new_ptr(int num) {
	int offset = 0;
	for (int i = 0; i < num; ++i) {
		const int sz = g_res._new_new[offset] * g_res._new_new[offset + 1];
		offset += sz + 2;
	}
	return g_res._new_new + offset;
}

static anim_data_t *get_ref_ptr(int num_ref, int anim_count) {
	anim_data_t *ad = 0;
	if (num_ref < 0) {
		ad = &g_res._ref_ref_use_ptr3;
	} else if (num_ref > 0) {
		assert(num_ref < anim_count);
		ad = &g_res._ref_ref[num_ref];
	}
	return ad;
}

static const uint8_t *fetch_level_data(const char *filename, int *size) {
	int slot_tmp = fio_open(filename, FIO_READ, 1);
	assert(g_res.slot_count < MAX_RES_SLOTS);
	g_res.slot_table[g_res.slot_count] = slot_tmp;
	++g_res.slot_count;
	const uint8_t *data = fio_fetch(slot_tmp);
	if (size) {
		*size = fio_size(slot_tmp);
	}
	return data;
}

static void load_ref_ref__v2(const uint8_t *data, int size) {
	static const uint16_t move_delta_offsets_table[15] = {
		0, 88, 92, 180, 236, 292, 374, 462, 518, 570, 682, 768, 854, 898, 942
	};
	const int anim_count = size / 0x40;
	assert((size % 0x40) == 0);
	assert(anim_count <= ARRAYSIZE(g_res._ref_ref));
	for (int i = 0; i < anim_count; ++i) {
		anim_data_t *ad = &g_res._ref_ref[i];
		ad->unk0 = *data++;
		ad->lifes = *data++;
		ad->cycles = *data++;
		ad->unk3 = *data++;
		ad->unk4 = *data++;
		ad->unk5 = *data++;
		ad->unk6 = *data++;
		ad->init_sprite_num = *data++;
		ad->colliding_opcode = *data++;
		ad->logic_opcode = *data++;
		ad->sound_num = *data++;
		ad->rnd = *data++;
		ad->sprite_num = *data++;
		ad->sprite_flags = *data++;
		ad->anim_w = READ_LE_UINT16(data); data += 2;
		ad->anim_h = READ_LE_UINT16(data); data += 2;
		ad->score = READ_LE_UINT16(data); data += 2;
		ad->bounding_box_x1 = READ_LE_UINT16(data); data += 2;
		ad->bounding_box_x2 = READ_LE_UINT16(data); data += 2;
		ad->bounding_box_y1 = READ_LE_UINT16(data); data += 2;
		ad->bounding_box_y2 = READ_LE_UINT16(data); data += 2;
		const int index = READ_LE_UINT16(data); data += 2;
		ad->move_data_ptr = (index != 0) ? move_delta_table + move_delta_offsets_table[index - 1] : 0;
		data += 2;
		ad->distance_dx = READ_LE_UINT16(data); data += 2;
		ad->distance_dy = READ_LE_UINT16(data); data += 2;
		ad->anim_data1_ptr = get_ref_ptr((int16_t)READ_LE_UINT16(data), anim_count); data += 2;
		data += 2;
		ad->anim_data2_ptr = get_ref_ptr((int16_t)READ_LE_UINT16(data), anim_count); data += 2;
		data += 2;
		ad->anim_data3_ptr = get_ref_ptr((int16_t)READ_LE_UINT16(data), anim_count); data += 2;
		data += 2;
		ad->dx = READ_LE_UINT16(data); data += 2;
		ad->dy = READ_LE_UINT16(data); data += 2;
		ad->anim_data4_ptr = get_ref_ptr((int16_t)READ_LE_UINT16(data), anim_count); data += 2;
		data += 2;
		ad->dx2 = READ_LE_UINT16(data); data += 2;
		ad->dy2 = READ_LE_UINT16(data); data += 2;
		ad->draw_dx = READ_LE_UINT16(data); data += 2;
		ad->draw_dy = READ_LE_UINT16(data); data += 2;
	}
}

static const uint16_t decor_pal_table__v3[16 * 5] = {
	0x000, 0xFED, 0xCD1, 0x7B2, 0x470, 0x231, 0x000, 0xFC3,
	0x07B, 0xB30, 0x3BF, 0x26A, 0x136, 0xF94, 0x853, 0x511,
	0x000, 0xFED, 0xABB, 0x789, 0x556, 0x434, 0x000, 0xFC3,
	0x07B, 0xB30, 0x3BF, 0x26A, 0x136, 0xF94, 0xB64, 0x844,
	0x000, 0xFED, 0xABB, 0x789, 0x556, 0x434, 0x000, 0xFC3,
	0x07B, 0xB30, 0x3BF, 0x26A, 0x136, 0xF94, 0xB64, 0x844,
	0x000, 0xFEC, 0xF00, 0x7BB, 0x9C5, 0xC97, 0x634, 0xDD3,
	0x07B, 0xA75, 0x045, 0x079, 0x493, 0x655, 0x221, 0xEB9,
	0x000, 0xFFF, 0xF00, 0x78D, 0x8C3, 0x642, 0xFC9, 0xED0,
	0x07B, 0x878, 0x221, 0x056, 0x086, 0x420, 0xA92, 0xFF7
};

static const uint16_t lifebar_pal[16] = {
	0x000, 0xFFF, 0x000, 0xF00, 0xF10, 0xF20, 0xF30, 0xF40,
	0xF50, 0xF60, 0xF70, 0xF80, 0xF90, 0xFA0, 0xFB0, 0xFC0
};

static const uint16_t digits_pal[16] = {
	0x000, 0xFFF, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000,
	0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000, 0x000
};

static const uint16_t carlos_pal[16] = {
	0x000, 0x18E, 0xFB9, 0xB98, 0xA98, 0x977, 0x666, 0x856,
	0xC74, 0xD94, 0x543, 0xA01, 0xFD0, 0xF80, 0x330, 0xFFF
};

static const uint16_t *degra_dat[] = {
	DEGRA00_COP,
	DEGRA01_COP,
	DEGRA03_COP,
	DEGRA04_COP,
	DEGRA50_COP
};

static void set_colors(int level) {
	static const int DEGRA_SIZE = 160;

	sys_set_copper_pal(degra_dat[level], DEGRA_SIZE);
	if (IS_GBA) {
		sys_set_palette_spr(&decor_pal_table__v3[level * 16], 16, 0);
		sys_set_palette_spr(carlos_pal, 16, 1);
		sys_set_palette_spr(digits_pal, 16, 2);
		sys_set_palette_spr(lifebar_pal, 16, 3);
	}
}

void res_load_level_data__v3(int level) {
	int slot_tmp, size;
	char filename[32];
	const uint8_t *file_data;
	const uint16_t *pal_data = &decor_pal_table__v3[level * 16];

	res_unload_level_data();

	/* tiles gfx */
	static const uint8_t blk[5] = { 1, 2, 2, 4, 5 };
	snprintf(filename, sizeof(filename), "AA%dBLK", blk[level]);
	load_tile(filename, level, pal_data);

	/* tiles flags */
	snprintf(filename, sizeof(filename), "AA%dREF.%s", level + 1, g_res.dat_ext);
	slot_tmp = fio_open(filename, FIO_READ, 1);
	size = fio_size(slot_tmp);
	if (size != DECOR_REF_SIZE) {
		error("Unexpected file size '%s' %d bytes", filename, size);
	} else {
		fio_read(slot_tmp, g_res._decor_ref, size);
	}
	fio_close(slot_tmp);

	/* tile map (will be modified, make a copy) */
	snprintf(filename, sizeof(filename), "A%d0CDG.%s", level + 1, g_res.dat_ext);
	slot_tmp = fio_open(filename, FIO_READ, 1);
	size = fio_size(slot_tmp);
	if (size > DECOR_CDG_SIZE) {
		error("Unexpected file size '%s' %d bytes", filename, size);
	} else {
		fio_read(slot_tmp, g_res._decor_cdg, size);
	}
	fio_close(slot_tmp);

	/* workaround level 2 vertical bars */
	if (level == 1) {
		static const int16_t black_bars_x[] = { 258, 267, 271, -1 };
		for (int i = 0; black_bars_x[i] != -1; ++i) {
			const int offset = black_bars_x[i] * _screen_cdg_tile_map_h + 45;
			if (memcmp(g_res._decor_cdg + offset, "\x19\x19\x19\x19", 4) == 0) {
				memcpy(g_res._decor_cdg + offset, "\xBF\xA8\xA8\xA8", 4);
			}
		}
	}

	/* tiles update data */
	snprintf(filename, sizeof(filename), "A%d0NEW.%s", level + 1, g_res.dat_ext);
	g_res._new_new = fetch_level_data(filename, &size);

	/* carlos sprites */
	if (IS_GBA) {
		snprintf(filename, sizeof(filename), "S00A.%s", g_res.gfx_ext);
		g_res._carlos1_spr = load_sprite(filename, level, pal_data);
		snprintf(filename, sizeof(filename), "S01A.%s", g_res.gfx_ext);
		g_res._carlos2_spr = load_sprite(filename, level, pal_data);
	} else {
		static const uint8_t s0x[] = { 0, 2, 2, 6, 8 };
		snprintf(filename, sizeof(filename), "S0%d.%s", s0x[level], g_res.gfx_ext);
		g_res._carlos1_spr = load_sprite(filename, level, pal_data);
		snprintf(filename, sizeof(filename), "S0%d.%s", s0x[level] + 1, g_res.gfx_ext);
		g_res._carlos2_spr = load_sprite(filename, level, pal_data);
	}

	/* monster/level sprites */
	snprintf(filename, sizeof(filename), "S%d.%s", 11 + level, g_res.gfx_ext);
	g_res._level_spr = load_sprite(filename, level, pal_data);

	/* boss sprites */
	snprintf(filename, sizeof(filename), "S%d.%s", 21 + level, g_res.gfx_ext);
	g_res._boss_spr = load_sprite(filename, level, pal_data);

	/* oscar sprites */
	snprintf(filename, sizeof(filename), "S%d.%s", 31 + level, g_res.gfx_ext);
	g_res._oscar_spr = load_sprite(filename, level, pal_data);

	/* friend sprites */
	snprintf(filename, sizeof(filename), "S%d.%s", 41 + level, g_res.gfx_ext);
	g_res._friend_spr = load_sprite(filename, level, pal_data);

	/* monster flags */
	snprintf(filename, sizeof(filename), "A%d0POS.%s", level + 1, g_res.dat_ext);
	g_res._posit_ref = fetch_level_data(filename, &size);
	assert((size % 10) == 0);
	g_res._posit_count = size / 10;

	/* monster anim_data */
	snprintf(filename, sizeof(filename), "REF%dREF.%s", level + 1, g_res.dat_ext);
	slot_tmp = fio_open(filename, FIO_READ, 1);
	file_data = fio_fetch(slot_tmp);
	load_ref_ref__v2(file_data, fio_size(slot_tmp));
	fio_close(slot_tmp);

	_screen_cdg_tile_map_h = READ_LE_UINT16(g_res._posit_ref + 2);
	_screen_cdg_tile_map_w = READ_LE_UINT16(g_res._posit_ref + 4);
	_screen_cdg_tile_map_start_y = READ_LE_UINT16(g_res._posit_ref + 6);
	_screen_cdg_tile_map_start_x = READ_LE_UINT16(g_res._posit_ref + 8);
	--g_res._posit_count;
	g_res._posit_ref += 10;
	if (READ_LE_UINT16(g_res._posit_ref) == 0xFFFE) {
		_screen_cdg_tile_map_h2 = READ_LE_UINT16(g_res._posit_ref + 2);
		const int w = READ_LE_UINT16(g_res._posit_ref + 4);
		res_load_fond(w & 0x7F, level);
		_screen_cdg_unk2 = w & 0x80;
		--g_res._posit_count;
		g_res._posit_ref += 10;
	}

	sys_set_tilemap_data(g_res._decor_cdg, _screen_cdg_tile_map_w, _screen_cdg_tile_map_h, _screen_cdg_offsets);

	if (g_res.converted_data || IS_GBA) {
		snprintf(filename, sizeof(filename), "LIFEBAR.%s", g_res.gfx_ext);
		g_res._lifebar_spr = load_sprite(filename, 0, pal_data);
		snprintf(filename, sizeof(filename), "DIGITS.%s", g_res.gfx_ext);
		g_res._digits_spr = load_sprite(filename, 0, pal_data);
	} else {
		g_res._lifebar_spr = load_static_sprite(LIFEBAR_SPR, lifebar_pal);
		g_res._digits_spr = load_static_sprite(DIGITS_SPR, digits_pal);
	}

	set_colors(level);

	g_res.level_data_loaded = true;
}

void res_unload_level_data() {
	if (g_res.level_data_loaded) {
		for (int i = 0; i < g_res.slot_count; ++i) {
			fio_close(g_res.slot_table[i]);
		}
		g_res.slot_count = 0;
		g_res.level_data_loaded = false;
	}
	for (int i = 0; i < MAX_RES_SLOTS; ++i) {
		g_res.slot_table[i] = -1;
	}
	sys_clear_sprite_data();
}

void res_load_fond(int num, int level) {
	load_fond(num, level, &decor_pal_table__v3[level * 16]);
}

void res_init_static_data__v3() {
	memset(&g_res.anim_data0, 0, sizeof(g_res.anim_data0));
	g_res.anim_data0.unk0 = -1;
	g_res.anim_data0.sound_num = -1;
}
