
#include <ctype.h>
#include "bitmap.h"
#include "decode.h"
#include "fileio.h"
#include "game.h"
#include "resource.h"
#include "systemstub.h"

static const int MAX_SPRITES = 192;

static uint8_t *convert_tile_ega(const uint8_t *data, const uint16_t *pal) {
	static uint8_t buffer[256 * 16 * 16];
	static const int COUNT = 256;
	static const int W = 16;
	static const int H = 16;
	for (int i = 0; i < COUNT; ++i) {
		decode_tile(buffer + i * W, W * COUNT, data);
		data += 8 * 16;
	}
	return bitmap_from_pal_data(buffer, W * COUNT, H, pal, 0);
}

static uint8_t *convert_sprite_ega(const uint8_t *data, const uint16_t *pal, int *count, sys_rect_t *r) {
	int offsets[MAX_SPRITES];
	int w, h;
	uint8_t *p = 0;
	*count = READ_LE_UINT16(data) / 2;
	assert(*count <= MAX_SPRITES);
	w = h = 0;
	for (int i = 0; i < *count; ++i) {
		const int offset = READ_LE_UINT16(data + i * 2);
		const int spr_w = READ_LE_UINT16(data + offset);
		const int spr_h = READ_LE_UINT16(data + offset + 2);
		r[i].x = 0;
		r[i].y = h;
		r[i].w = spr_w;
		r[i].h = spr_h;
		h += spr_h;
		if (w < spr_w) {
			w = spr_w;
		}
		offsets[i] = offset;
	}
	uint8_t *buffer = (uint8_t *)calloc(w * h, 1);
	if (!buffer) {
		error("Failed to allocate %d bytes for sprite buffer", w * h);
	} else {
		int offset = 0;
		for (int i = 0; i < *count; ++i) {
			const uint8_t *src = data + offsets[i];
			const int spr_w = READ_LE_UINT16(src);
			const int spr_h = READ_LE_UINT16(src + 2);
			decode_sprite_ega(buffer + offset, w, src + 4, spr_w, spr_h);
			offset += spr_h * w;
		}
		p = bitmap_from_pal_data(buffer, w, h, pal, 0);
		free(buffer);
	}
	return p;
}

static uint8_t *convert_image_ega(const uint8_t *data, const uint16_t *pal, int w, int h) {
	uint8_t *bmp = 0;
	uint8_t *buffer = (uint8_t *)calloc(w * h, 1);
	if (!buffer) {
		error("Failed to allocate %d bytes for image buffer", w * h);
	} else {
		decode_image_ega(buffer, w, data, w, h);
		bmp = bitmap_from_pal_data(buffer, w, h, pal, -1);
		free(buffer);
	}
	return bmp;
}

static int read_file(const char *name, void *buffer, int max_size) {
	int slot, size = 0;

	slot = fio_open(name, FIO_READ, 0);
	if (slot >= 0) {
		size = fio_size(slot);
		if (size > max_size) {
			error("Unexpected file size '%s' %d bytes", name, size);
		} else {
			fio_read(slot, buffer, size);
		}
		fio_close(slot);
	}
	return size;
}

static int read_sprite_offsets(const char *str, sys_rect_t *r) {
	int i, count, state = 0;

	for (i = 0; str && *str; ) {
		if (state == 0) {
			if (sscanf(str, "%d", &count) != 1) {
				break;
			}
			state = 1;
		} else {
			if (sscanf(str, "%d %d %d %d", &r[i].x, &r[i].y, &r[i].w, &r[i].h) != 4) {
				break;
			}
			++i;
		}
		str = strpbrk(str, "\r\n");
		if (str) {
			while (*str && isspace(*str)) {
				++str;
			}
		}
	}
	return i;
}

static const char *get_amiga_sprite_name(const char *name) {
	if (name[1] == '0') { /* 'S0x' carlos */
		const int num = name[2] - '0';
		return ((num & 1) != 0) ? "S01A" : "S00A";
	}
	if (name[1] == '3') { /* 'S3x' oscar */
		return "S31A";
	}
	return name;
}

int load_sprite(const char *name, int level, const uint16_t *pal)  {
	const uint8_t *data;
	int slot, spr, len, count = 0;
	sys_rect_t dim[MAX_SPRITES];

	if (g_res.converted_data) {
		char buffer[RES_TXT_SIZE], filename[32];
		char *ext;

		/* offsets */
		snprintf(filename, sizeof(filename), "%s", name);
		ext = strrchr(filename, '.');
		if (ext) {
			strcpy(ext + 1, "TXT");
		}
		len = read_file(filename, buffer, RES_TXT_SIZE);
		if (len > 0) {
			buffer[len] = 0;
			count = read_sprite_offsets(buffer, dim);
		}
		/* bitmaps */
		if (g_res.gfx_scale_w != 1) {
			assert(g_res.gfx_scale_w == g_res.gfx_scale_h);
			snprintf(filename, sizeof(filename), "%dx/%s.%s", g_res.gfx_scale_w, get_amiga_sprite_name(name), g_res.gfx_ext);
		} else {
			snprintf(filename, sizeof(filename), "%s.%s", get_amiga_sprite_name(name), g_res.gfx_ext);
		}
		slot = fio_open(filename, FIO_READ, 1);
		data = fio_fetch(slot);
		spr = sys_load_sprite_data(data, fio_size(slot), count, dim);
		fio_close(slot);
	} else {
		uint8_t *bmp;

		slot = fio_open(name, FIO_READ, 1);
		data = fio_fetch(slot);
		bmp = convert_sprite_ega(data, pal, &count, dim);
		if (bmp) {
			len = READ_LE_UINT32(bmp + 2);
			spr = sys_load_sprite_data(bmp, len, count, dim);
			free(bmp);
		} else {
			spr = -1;
		}
		fio_close(slot);
	}
	return spr;
}

int load_static_sprite(const uint8_t *data, const uint16_t *pal) {
	int count, len, spr = -1;
	uint8_t *bmp;
	sys_rect_t dim[MAX_SPRITES];

	bmp = convert_sprite_ega(data, pal, &count, dim);
	if (bmp) {
		len = READ_LE_UINT32(bmp + 2);
		spr = sys_load_sprite_data(bmp, len, count, dim);
		free(bmp);
	}
	return spr;
}

void load_tile(const char *name, int level, const uint16_t *pal) {
	char filename[64];
	int slot, len;
	const uint8_t *data;

	if (g_res.gfx_scale_w != 1) {
		assert(g_res.gfx_scale_w == g_res.gfx_scale_h);
		snprintf(filename, sizeof(filename), "%dx/%s.%s", g_res.gfx_scale_w, name, g_res.gfx_ext);
	} else {
		snprintf(filename, sizeof(filename), "%s.%s", name, g_res.gfx_ext);
	}
	debug(DBG_RESOURCE, "sys_set_tile_data %s name '%s'", filename, name);

	slot = fio_open(filename, FIO_READ, 1);
	data = fio_fetch(slot);
	if (g_res.converted_data) {
		sys_set_tile_data(data, fio_size(slot), 0);
	} else {
		uint8_t *bmp = convert_tile_ega(data, pal);
		if (bmp) {
			len = READ_LE_UINT32(bmp + 2);
			sys_set_tile_data(bmp, len, 0);
			free(bmp);
		}
	}
	fio_close(slot);
}

void load_fond(int num, int level, const uint16_t *pal) {
	static const int W = 640;
	static const int H = 32;
	char filename[32];
	if (g_res.gfx_scale_w != 1) {
		assert(g_res.gfx_scale_w == g_res.gfx_scale_h);
		snprintf(filename, sizeof(filename), "%dx/FOND%02d.%s", g_res.gfx_scale_w, num, g_res.gfx_ext);
	} else {
		snprintf(filename, sizeof(filename), "FOND%02d.%s", num, g_res.gfx_ext);
	}
	int slot = fio_open(filename, FIO_READ, 0);
	if (slot >= 0) {
		const uint8_t *file_data = fio_fetch(slot);
		if (g_res.converted_data) {
			sys_set_background_frame(file_data, fio_size(slot));
		} else {
			uint8_t *bmp = convert_image_ega(file_data, pal, W, H);
			if (bmp) {
				const int len = READ_LE_UINT32(bmp + 2);
				sys_set_background_frame(bmp, len);
				free(bmp);
			}

		}
		fio_close(slot);
	} else {
		sys_set_background_frame(0, 0);
	}
}

static const uint16_t carlos_image_pal_table[16 * 9] = {
	/* I20 */
	0x000, 0xC9A, 0x128, 0x069, 0x04E, 0x339, 0x549, 0x659,
	0xF50, 0x520, 0xEE0, 0xDBA, 0xA7B, 0xC9B, 0xECA, 0xFFF,
	/* I21 */
	0x000, 0xC9A, 0x238, 0x069, 0x439, 0x659, 0x75A, 0xA6B,
	0xB7A, 0x128, 0xDBA, 0xEBA, 0xECA, 0xEE8, 0xEF0, 0xFFF,
	/* I22 */
	0x000, 0xC9A, 0x238, 0x339, 0x548, 0x739, 0x86A, 0xA6B,
	0xF00, 0xB8B, 0x083, 0xD9A, 0xDBA, 0xEBA, 0xECA, 0xFFF,
	/* I23 */
	0x000, 0xC9B, 0x238, 0x069, 0x239, 0x549, 0x659, 0xA6B,
	0xC88, 0xDAA, 0xC5D, 0x083, 0xECA, 0xEE8, 0xEF0, 0xFFF,
	/* I24 */
	0x000, 0xC9A, 0x339, 0x549, 0x04E, 0x659, 0x75A, 0xA6B,
	0xF50, 0x238, 0xDA9, 0xEBA, 0xB7B, 0xC9B, 0xECA, 0xFFF,
	/* I25 */
	0x000, 0xB8B, 0x083, 0x521, 0x128, 0x04B, 0x339, 0x749,
	0xF00, 0xE86, 0xECA, 0xE4F, 0x005, 0xEE8, 0xEF0, 0xFFF,
	/* I00 */
	0x000, 0x8CE, 0xA2A, 0xC4C, 0xE8E, 0xEAE, 0x28E, 0x4AE,
	0x808, 0xE20, 0xE60, 0xEA0, 0xEC8, 0x066, 0x044, 0xEEE,
	/* I01 */
	0x000, 0xC9A, 0x238, 0x659, 0xA6B, 0x016, 0xC9B, 0xDAA,
	0xF00, 0xF50, 0xEB9, 0xEBA, 0xECA, 0xEE8, 0xEF0, 0xFFF,
	/* I02 */
	0x000, 0xDAA, 0x005, 0x127, 0x238, 0x449, 0x659, 0xA7A,
	0xF00, 0xF84, 0x620, 0xEBA, 0xECA, 0xEE8, 0xEF0, 0xFFF
};

void res_load_img(int num) {
	static const int W = 320;
	static const int H = 200;
	int slot;
	char filename[16];

	snprintf(filename, sizeof(filename), "I%02d.%s", num, g_res.gfx_ext);
	slot = fio_open(filename, FIO_READ, 0);
	if (slot >= 0) {
		sys_set_screen_mode(SSM_BITMAP);
		const uint8_t *file_data = fio_fetch(slot);
		if (g_res.converted_data) {
			sys_set_bitmap_data(file_data, fio_size(slot));
		} else {
			const uint16_t *pal = 0;
			if (num >= 0 && num <= 2) {
				pal = &carlos_image_pal_table[(num + 6) * 16];
			} else if (num >= 20 && num <= 25) {
				pal = &carlos_image_pal_table[(num - 20) * 16];
			}
			if (pal) {
				uint8_t *bmp = convert_image_ega(file_data, pal, W, H);
				if (bmp) {
					const int len = READ_LE_UINT32(bmp + 2);
					sys_set_bitmap_data(bmp, len);
					free(bmp);
				}
			} else {
				warning("No palette for img %d", num);
			}
		}
		fio_close(slot);
		g_res.img = num;
	} else {
		sys_set_bitmap_data(0, 0);
		g_res.img = -1;
	}
}

