
#include "gba_compression.h"
#include "gba_dma.h"
#include "gba_input.h"
#include "gba_video.h"
#include "gba_sprites.h"
#include "gba_interrupt.h"
#include "gba_systemcalls.h"
#include "maxmod.h"

#include "fileio.h"
#include "systemstub.h"

#define MAX_SPRITES 128

#define HITMODE_PAL_NUM 15

#define PAL_COLORS 16

typedef struct {
	const uint8_t *spr_data;
	int spr_size;
	int tiles_count;
	uint16_t attr[3];
} oam_entry_t;

typedef struct {
	const uint8_t *data;
	int size;
} sprite_data_t;

int sys_exit_flag;
static bool palette_changed;
static int released_key_mask, pressed_key_mask;
static sys_screen_mode_e current_mode;
static int screen_w, screen_h;
static uint16_t tiles_palette[16];
static uint16_t copper_palette[161];
static const uint8_t *tilemap_data, *tilemap_lut;
static int tilemap_w, tilemap_h;
static int tilemap_origin_x, tilemap_origin_y;
static sprite_data_t sprites_data_table[MAX_SPRITES] GCC_EXT;
static int sprites_data_count;
static int sprites_count, sprites_count_prev;
static int obj_tiles_size;
static const uint8_t *tile_data;
static bool tile_data_changed;
static oam_entry_t oam_table[MAX_SPRITES];
static const uint16_t *copper_pal_data;
static bool copper_pal_changed;
static int screen_voffset;

static const uint8_t gamma_lut[32] = { /* pow(i / 31., .8) * 31. */
	0x00, 0x01, 0x03, 0x04, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10, 0x11,
	0x12, 0x13, 0x14, 0x14, 0x15, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x1A, 0x1B, 0x1C, 0x1D, 0x1E, 0x1F
};

static void vblank_interrupt() {
	/* reset the DMA */
	REG_DMA0CNT = 0;
	if (current_mode == SSM_TILED) {
		DMA0COPY(&copper_palette[1], &BG_COLORS[0], DMA_DST_FIXED | DMA_SRC_INC | DMA_REPEAT | DMA16 | DMA_HBLANK | 1);
		/* setup first color, hblank occurs at the end of line */
		BG_COLORS[0] = copper_palette[0];
	}
}

static void sound_init() {
	const int slot = fio_open("SNDBANK", FIO_READ, 0);
	void *data = (void *)fio_fetch(slot);
	static const int channels = 4;
	mmInitDefault(data, channels);
}

void sys_init(int w, int h, int sw, int sy, const char *title) {
	screen_w = w;
	screen_h = h;
	/* 'white' blending ; set attr[0] |= OBJ_MODE(1) when !HITMODE */
/*	REG_BLDCNT = (1 << 8) | (2 << 6) | (1 << 4);
	REG_BLDALPHA = (16 << 0);
	REG_BLDY = (8 << 0);*/
	/* 'black' blending ; set attr[0] |= OBJ_MODE(1) when HITMODE */
/*	REG_BLDCNT = (1 << 8) | (1 << 6);
	REG_BLDALPHA = (8 << 0);*/
	/* initialize sound */
	sound_init();
	/* setup vblank interrupt handler */
	irqInit();
	irqSet(IRQ_VBLANK, mmVBlank);
	irqEnable(IRQ_VBLANK);
	mmSetVBlankHandler(vblank_interrupt);
}

void sys_fini() {
}

static uint16_t convert_amiga_color(uint16_t color) {
	uint8_t r, g, b;

	r = (color >> 8) & 15;
	g = (color >> 4) & 15;
	b =  color       & 15;
	r = gamma_lut[(r << 1) | (r >> 3)];
	g = gamma_lut[(g << 1) | (g >> 3)];
	b = gamma_lut[(b << 1) | (b >> 3)];
	return RGB5(r, g, b);
}

static void convert_copper_pal(const uint16_t *pal_data) {
	int i;

	for (i = 0; i < 160; ++i) {
		copper_palette[i] = convert_amiga_color(pal_data[i]);
	}
}

void sys_set_screen_mode(sys_screen_mode_e mode) {
	int i;

	current_mode = mode;
	switch (current_mode) {
	case SSM_BITMAP:
		REG_DISPCNT = MODE_3 | BG2_ON;
		break;
	case SSM_TILED:
		/* CHAR(0,1) : decor tiles 0x8000 bytes : 16 * 16 * 256 / 2 (4bpp) */
		/* CHAR(2) : background frame tiles */
		REG_BG0CNT = BG_16_COLOR | CHAR_BASE(0) | SCREEN_BASE(31) | BG_SIZE_0;
		REG_BG1CNT = BG_16_COLOR | CHAR_BASE(2) | SCREEN_BASE(30) | BG_SIZE_0;
		REG_DISPCNT = MODE_0 | OBJ_1D_MAP | OBJ_ON | BG0_ON;
		for (i = 0; i < 16; ++i) {
			OBJ_COLORS[HITMODE_PAL_NUM * 16 + i] = RGB5(31, 31, 31);
		}
		for (i = 0; i < MAX_SPRITES; ++i) {
			OAM[i].attr0 = OBJ_DISABLE;
		}
		break;
	}
}

void sys_set_palette_spr(const uint16_t *pal_data, int num_colors, int pal_num) {
	int i;

	assert(num_colors == PAL_COLORS);
	if (pal_num == 0) {
		/* sprites and tiles share the same palette colors */
		for (i = 0; i < PAL_COLORS; ++i) {
			const uint16_t color = convert_amiga_color(pal_data[i]);
			OBJ_COLORS[i] = tiles_palette[i] = color;
		}
		palette_changed = true;
	} else {
		const int offset = pal_num * PAL_COLORS;
		/* extra sprites (Amiga) */
		for (i = 0; i < PAL_COLORS; ++i) {
			const uint16_t color = convert_amiga_color(pal_data[i]);
			OBJ_COLORS[offset + i] = color;
		}
	}
}

static void clear_palette() {
	uint32_t color = 0;
	CpuFastSet(&color, BG_COLORS, FILL | COPY32 | (256 * sizeof(uint16_t) / 4));
}

void sys_fade_in_palette() {
	clear_palette();
}

void sys_fade_out_palette() {
}

static void update_tilemap(int x, int y) {
	uint16_t *dstmap = (uint16_t *)MAP_BASE_ADR(31);
	const uint8_t *srcmap;
	int i, j;

	y += screen_voffset;

	REG_BG0HOFS = x & 0xF;
	REG_BG0VOFS = y & 0xF;
	x >>= 4;
	y >>= 4;

	srcmap = &tilemap_data[x * tilemap_h + y];
	for (j = 0; j < 16 * 32; j += 32) {
		for (i = 0; i < 16 * tilemap_h; i += tilemap_h) {
			const uint16_t tile_num = tilemap_lut[srcmap[i]] << 2;
			*(uint32_t *)(dstmap + j)      =  tile_num      | ((tile_num + 1) << 16);
			*(uint32_t *)(dstmap + j + 32) = (tile_num + 2) | ((tile_num + 3) << 16);
			dstmap += 2;
		}
		++srcmap;
	}
}

static void sync_display_tiled() {
	uint8_t *dst;
	int i, tiles_count;

	/* update palette */
	if (palette_changed) {
		CpuFastSet(tiles_palette, BG_COLORS, COPY32 | (16 * sizeof(uint16_t) / 4));
		palette_changed = false;
	}

	/* update copper data */
	if (copper_pal_changed) {
		convert_copper_pal(copper_pal_data);
		copper_pal_changed = false;
	}

	/* update tilemap data */
	if (tile_data_changed) {
		LZ77UnCompVram(tile_data, CHAR_BASE_ADR(0));
		tile_data_changed = false;
	}

	/* update tilemap */
	update_tilemap(tilemap_origin_x, tilemap_origin_y);

	/* setup sprites, in reverse order */
	tiles_count = 0;
	dst = OBJ_BASE_ADR;
	const oam_entry_t *oe = &oam_table[sprites_count - 1];
	for (i = 0; i < sprites_count; ++i, --oe) {
		OAM[i].attr0 = oe->attr[0];
		OAM[i].attr1 = oe->attr[1];
		OAM[i].attr2 = oe->attr[2] | OBJ_CHAR(tiles_count);
		CpuFastSet(oe->spr_data, dst, COPY32 | (oe->spr_size / 4));
		tiles_count += oe->tiles_count;
		dst += oe->tiles_count * 32;
	}

	/* hide previously setup sprites */
	for (i = sprites_count; i < sprites_count_prev; ++i) {
		OAM[i].attr0 = OBJ_DISABLE;
	}

	sprites_count_prev = sprites_count;
}

void sys_set_screen_offset(int offset) {
	screen_voffset = offset;
}

void sys_fill_rect(const sys_rect_t *rect, int r, int g, int b) {
	/* unused on GBA, sprite is used for lifebar */
}

void sys_update_screen() {
	mmFrame();
	VBlankIntrWait();
	scanKeys();
	if (current_mode == SSM_TILED) {
		sync_display_tiled();
	}
	mmFrame();
	VBlankIntrWait();
	scanKeys();
}

void sys_set_bitmap_data(const uint8_t *bmp_data, int size) {
	VBlankIntrWait();
	LZ77UnCompVram(bmp_data, (void *)VRAM);
}

void sys_set_tile_data(const uint8_t *data, int size, uint8_t color) {
	tile_data = data;
	tile_data_changed = true;
}

void sys_set_tilemap_data(const uint8_t *data, int w, int h, const uint8_t *anim_lut) {
	tilemap_data = data;
	tilemap_lut = anim_lut;
	tilemap_origin_x = 0;
	tilemap_origin_y = 0;
	tilemap_w = w;
	tilemap_h = h;
}

void sys_set_tilemap_origin(int x, int y) {
	tilemap_origin_x = x;
	tilemap_origin_y = y;
}

void sys_set_copper_pal(const uint16_t *pal_data, int num_lines) {
	copper_pal_data = pal_data;
	copper_pal_changed = true;
}

void sys_set_background_frame(const uint8_t *data, int size) {
	if (data) {
		const uint16_t tile_size = *(uint16_t *)data; data += 4;
		CpuFastSet(data, CHAR_BASE_ADR(2), COPY32 | (tile_size / 4));
		LZ77UnCompVram(data + tile_size, MAP_BASE_ADR(30));
		REG_BG1HOFS = 0;
		REG_BG1VOFS = 0;
		REG_DISPCNT |= BG1_ON;
	} else {
		REG_DISPCNT &= ~BG1_ON;
	}
}

void sys_clear_sprite_data() {
	sprites_data_count = 0;
}

int sys_load_sprite_data(const uint8_t *data, int size, int count, sys_rect_t *r) {
	int spr_num = -1;

	if (sprites_data_count < MAX_SPRITES) {
		spr_num = sprites_data_count;
		sprite_data_t *spr = &sprites_data_table[sprites_data_count];
		spr->data = data;
		spr->size = size;
		++sprites_data_count;
	}
	return spr_num;
}

void sys_get_sprite_dim(int spr_data, int num, int *w, int *h) {
	if (spr_data < sprites_data_count) {
		const uint8_t *data = sprites_data_table[spr_data].data + 4;
		const uint32_t offset = *(uint32_t *)(data + 4 * num);
		data += offset;
		*w = *(uint16_t *)(data);
		*h = *(uint16_t *)(data + 2);
	}
}

void sys_clear_sprite_list() {
	sprites_count = 0;
	obj_tiles_size = 0;
}

void sys_add_to_sprite_list(int spr_data_num, int num, int x, int y, int flags, int pal_num) {
	const uint8_t *spr_data = sprites_data_table[spr_data_num].data + 4;
	const uint32_t offset = *(uint32_t *)(spr_data + 4 * num);
	if (offset != 0) {
		int i, j;
		int obj_x, obj_y, cur_x, cur_y;
		int num_spr_h, num_spr_w;
		int tile_h, tile_w;
		int gba_mode;
		oam_entry_t *oe;

		spr_data += offset;
		num_spr_h = spr_data[5];
		num_spr_w = spr_data[6];
		spr_data += 8;

		if (flags & SSF_HITMODE) {
			pal_num = HITMODE_PAL_NUM;
		}

		for (j = 0, obj_y = y; j < num_spr_h; ++j, obj_y += 32) {
			for (i = 0, obj_x = x; i < num_spr_w; ++i, obj_x += 32) {
				tile_h = *(uint16_t *)spr_data; spr_data += 2;
				tile_w = *(uint16_t *)spr_data; spr_data += 2;
				if (obj_x + tile_w <= 0 || obj_x >= screen_w || obj_y + tile_h <= 0 || obj_y >= screen_h) {
					spr_data += 2;
					const int spr_size = *(uint16_t *)spr_data; spr_data += 2;
					spr_data += spr_size;
				} else {
					if (sprites_count >= ARRAYSIZE(oam_table)) {
						warning("sys_add_to_sprite_list() oam_table overflow");
						break;
					}
					if (obj_tiles_size >= 32 * 1024) {
						warning("sys_add_to_sprite_list() obj_tiles_size %d >= 32ko", obj_tiles_size);
						break;
					}
					cur_x = obj_x;
					if (cur_x < 0) {
						cur_x = 512 + cur_x;
					}
					cur_y = obj_y;
					if (cur_y < 0) {
						cur_y = 256 + cur_y;
					}

					oe = &oam_table[sprites_count];
					++sprites_count;

					gba_mode = *spr_data++;
					oe->tiles_count = *spr_data++;
					oe->spr_size = *(uint16_t *)spr_data; spr_data += 2;
					oe->spr_data = spr_data;
					oe->attr[0] = OBJ_Y(cur_y) | OBJ_16_COLOR | OBJ_SHAPE((gba_mode >> 2));
					oe->attr[1] = OBJ_X(cur_x) | OBJ_SIZE((gba_mode & 3));
					oe->attr[2] = OBJ_PRIORITY(0) | OBJ_PALETTE(pal_num);
					spr_data += oe->spr_size;
					obj_tiles_size += oe->spr_size;
				}
			}
		}
	}
}

static int convert_key_mask(int mask) {
	static const struct {
		int gba_key_mask;
		int sys_key_mask;
	} key_mask_table[] = {
		{ KEY_LEFT, SKM_LEFT },
		{ KEY_RIGHT, SKM_RIGHT },
		{ KEY_UP, SKM_UP },
		{ KEY_DOWN, SKM_DOWN },
		{ KEY_START, SKM_PAUSE },
		{ KEY_SELECT, SKM_INVENTORY },
		{ KEY_A | KEY_B, SKM_ACTION },
		{ KEY_L, SKM_F1 },
		{ KEY_R, SKM_F2 }
	};
	int i, sys_mask = 0;

	for (i = 0; i < ARRAYSIZE(key_mask_table); ++i) {
		if (mask & key_mask_table[i].gba_key_mask) {
			sys_mask |= key_mask_table[i].sys_key_mask;
		}
	}
	return sys_mask;
}

static void update_key_mask() {
	released_key_mask = convert_key_mask(keysUp());
	pressed_key_mask &= ~released_key_mask;
	pressed_key_mask |= convert_key_mask(keysDown());
}

void sys_process_events() {
	update_key_mask();
}

int sys_get_keys_pressed_mask() {
	return pressed_key_mask;
}

int sys_get_keys_released_mask() {
	return released_key_mask;
}

void sys_play_sfx(int num, const uint8_t *data, int size) {
	mmEffect(num);
}

void sys_play_module(int num, const uint8_t *data, int size) {
	mmStart(num, MM_PLAY_LOOP);
}

void sys_stop_module() {
	mmStop();
}

void sys_print_string(const char *str) {
#ifdef CARLOS_DEBUG
	__asm volatile(
		"mov r0, %0;"
		"swi 0xFF;"
			: /* no output */
			: "r" (str)
			: "r0"
	);
#endif
}
