
#include <SDL.h>
#include <SDL_image.h>
#include <SDL_mixer.h>
#ifdef __EMSCRIPTEN__
#include <emscripten.h>
#endif
#include "systemstub.h"

#define TILES_SURFACE_W 512
#define TILES_SURFACE_H 256

#define MAX_SPRITES 128

#define TILE_W 16
#define TILE_H 16
#define MAX_TILES 256

#define MAX_SPRITES_SHEETS 10

#define MAX_FILL_RECTS 2

#define TICK_MUL 2000
#define TICK_DIV 60

#define ALPHA_STEP 16

typedef struct {
	int sheet;
	int num;
	int x, y;
	int flags;
} sprite_t;

typedef struct {
	int count;
	SDL_Rect r[256];
	SDL_Surface *surface;
} spritesheet_t;

typedef struct {
	uint8_t *ptr;
	int size;
	SDL_Surface *surface;
} offscreen_buffer_t;

typedef struct {
	int a, r, g, b;
} color_t;

typedef struct {
	SDL_Rect r;
	color_t color;
} fill_rect_t;

int sys_exit_flag;

static color_t palette[256];
static uint16_t screen_palette[256];
static SDL_PixelFormat *screen_fmt;
static SDL_Surface *screen_surface, *fade_surface;
static SDL_Surface *background_surface;
static offscreen_buffer_t copper;
static spritesheet_t sprites_sheets_table[MAX_SPRITES_SHEETS];
static int sprites_sheets_count;
static int tilemap_origin_x, tilemap_origin_y, tilemap_x, tilemap_y;
static int released_key_mask, pressed_key_mask;
static uint8_t *tiles_buffer;
static const uint8_t *tilemap_lut, *tilemap_data;
static int tilemap_w, tilemap_h;
static int screen_w, screen_h;
static int scale_w, scale_h;
static sys_screen_mode_e current_mode;
static bool fullscreen_flag;
static bool pause_flag;
static uint32_t last_frame_time_stamp;
static sprite_t sprite_list[MAX_SPRITES];
static int sprite_list_size;
static int screen_voffset;
static fill_rect_t fill_rects_table[MAX_FILL_RECTS];
static int fill_rects_count;
static int cur_alpha, dst_alpha;
static Mix_Chunk *sound;
static Mix_Music *music;

static SDL_Surface *load_graphics(const uint8_t *data, int size, bool transparent);

static void setup_screen_surface(bool fullscreen) {
	int w, h;

	fullscreen_flag = fullscreen;
	w = screen_w * scale_w;
	h = screen_h * scale_h;
	screen_surface = SDL_SetVideoMode(w, h, 16, fullscreen_flag ? (SDL_FULLSCREEN | SDL_HWSURFACE) : SDL_HWSURFACE);
	if (!screen_surface) {
		error("Unable to allocate screen_surface");
	}
	screen_fmt = screen_surface->format;
	fade_surface = SDL_CreateRGBSurface(SDL_SWSURFACE, w, h, 16, screen_fmt->Rmask, screen_fmt->Gmask, screen_fmt->Bmask, screen_fmt->Amask);
	if (!fade_surface) {
		error("Unable to allocate fade_surface");
	}
	SDL_FillRect(fade_surface, 0, SDL_MapRGB(fade_surface->format, 0, 0, 0));
	SDL_SetAlpha(fade_surface, SDL_SRCALPHA, 255);
	cur_alpha = dst_alpha = 255;
}

static void change_screen_surface(bool fullscreen) {
	SDL_FreeSurface(fade_surface);
	SDL_FreeSurface(screen_surface);
	setup_screen_surface(fullscreen);
}

#ifdef __EMSCRIPTEN__
static void sys_process_event(SDL_Event *ev);

static int event_handler(void *userdata, SDL_Event *ev) {
	sys_process_event(ev);
	return 0;
}
#endif

void sys_init(int w, int h, int sw, int sh, const char *title) {
	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO);
	SDL_ShowCursor(SDL_DISABLE);
	SDL_WM_SetCaption(title, NULL);
	IMG_Init(IMG_INIT_PNG);
	Mix_Init(MIX_INIT_MOD);

	Mix_OpenAudio(22050, AUDIO_S16SYS, 2, 4096);

	pressed_key_mask = 0;
	sys_exit_flag = 0;

	assert(w < TILES_SURFACE_W && h < TILES_SURFACE_H);

	screen_w = w;
	screen_h = h;
	scale_w = sw;
	scale_h = sh;
	current_mode = SSM_BITMAP;
	memset(screen_palette, 0, sizeof(screen_palette));

	/* initialize screen surface */
	fullscreen_flag = false;
	setup_screen_surface(fullscreen_flag);

	/* initialize offscreen buffers */
	copper.size  = screen_w * scale_w * screen_h * scale_h;
	copper.ptr = (uint8_t *)malloc(copper.size);
	if (!copper.ptr) {
		error("Unable to allocate %d bytes", copper.size);
	} else {
		const int copper_w = screen_w * scale_w;
		const int copper_h = screen_h * scale_h;
		copper.surface = SDL_CreateRGBSurface(SDL_SWSURFACE, copper_w, copper_h, 32, 0xFF0000, 0xFF00, 0xFF, 0xFF000000); // ARGB
	}

	tiles_buffer = (uint8_t *)malloc(MAX_TILES * TILE_H * TILE_W);
	if (!tiles_buffer) {
		error("Unable to allocate tiles buffer");
	}

	last_frame_time_stamp = SDL_GetTicks();
	pause_flag = false;

#ifdef __EMSCRIPTEN__
	emscripten_SDL_SetEventHandler(event_handler, 0);
#endif
}

void sys_fini() {
	IMG_Quit();
	Mix_CloseAudio();
	Mix_Quit();
	free(copper.ptr);
	copper.ptr = 0;
	if (copper.surface) {
		SDL_FreeSurface(copper.surface);
		copper.surface = 0;
	}
	if (background_surface) {
		SDL_FreeSurface(background_surface);
		background_surface = 0;
	}
	free(tiles_buffer);
	SDL_FreeSurface(fade_surface);
	SDL_Quit();
}

void sys_set_screen_mode(sys_screen_mode_e mode) {
	current_mode = mode;
}

static uint32_t color_rgba(const SDL_PixelFormat *fmt, const color_t *color) {
	return SDL_MapRGBA(fmt, color->r, color->g, color->b, color->a);
}

static uint32_t color_rgb(const SDL_PixelFormat *fmt, const color_t *color) {
	return SDL_MapRGB(fmt, color->r, color->g, color->b);
}

static uint32_t convert_amiga_color(const SDL_PixelFormat *fmt, uint16_t rgb, color_t *color) {
	int r = (rgb >> 8) & 15;
	int g = (rgb >> 4) & 15;
	int b =  rgb       & 15;
	r |= r << 4;
	g |= g << 4;
	b |= b << 4;
	color->a = 255;
	color->r = r;
	color->g = g;
	color->b = b;
	return SDL_MapRGB(fmt, r, g, b);
}

static color_t lerp_color(int alpha, const color_t *color1, const color_t *color2) {
	color_t color;
	color.a = 255;
	color.r = (color1->r * alpha / 256) + (color2->r * (256 - alpha) / 256);
	color.g = (color1->g * alpha / 256) + (color2->g * (256 - alpha) / 256);
	color.b = (color1->b * alpha / 256) + (color2->b * (256 - alpha) / 256);
	return color;
}

static void draw_copper(SDL_Surface *surface, const uint16_t *colors, int count) {
	const int dst_pitch = surface->pitch / sizeof(uint32_t);
	uint32_t *dst = (uint32_t *)surface->pixels;
	const SDL_PixelFormat *fmt = surface->format;
	for (int y = 0; y < surface->h; ++y) {
#if 0
		color_t rgb;
		uint32_t color = convert_amiga_color(fmt, colors[y * count / surface->h], &rgb);
#else
		color_t rgb1, rgb2, rgb3;
		const int pos = ((y * count) << 10) / surface->h;
		const uint16_t color1 = colors[ pos >> 10 ];
		const uint16_t color2 = ((pos >> 10) + 1) < count ? colors[(pos >> 10) + 1] : colors[count - 1];
		const int alpha = ((1 << 10) - (pos & 0x3FF)) >> 2;
		convert_amiga_color(fmt, color1, &rgb1);
		convert_amiga_color(fmt, color2, &rgb2);
		rgb3 = lerp_color(alpha, &rgb1, &rgb2);
		uint32_t color = color_rgb(fmt, &rgb3);
#endif
		for (int x = 0; x < surface->w; ++x) {
			dst[x] = color;
		}
		dst += dst_pitch;
	}
}

void sys_set_palette_spr(const uint16_t *pal_data, int num_colors, int pal_num) {
	int i, offset = pal_num * 16;
	for (i = 0; i < num_colors; ++i) {
		const uint16_t color = pal_data[i];
		screen_palette[offset + i] = convert_amiga_color(screen_fmt, color, &palette[offset + i]);
	}
	if (offset == 0) {
		palette[0].a = 0;
	}
}

void sys_fade_in_palette() {
	cur_alpha = 0;
	dst_alpha = 255;
}

void sys_fade_out_palette() {
	cur_alpha = 255;
	dst_alpha = 0;
}

void sys_set_bitmap_data(const uint8_t *bmp_data, int size) {
	if (background_surface) {
		SDL_FreeSurface(background_surface);
		background_surface = 0;
	}
	if (bmp_data != 0) {
		background_surface = load_graphics(bmp_data, size, false);
	}
	current_mode = SSM_BITMAP;
}

static uint8_t lookup_tile(int x, int y) {
	const int tile_x = (tilemap_x + x) >> 4;
	const int tile_y = (tilemap_y + y) >> 4;
	const uint8_t num = tilemap_data[tile_x * tilemap_h + tile_y];
	return tilemap_lut[num];
}

static void update_screen_tiled() {
	SDL_Rect r;
	// background
	SDL_BlitSurface(copper.surface, 0, screen_surface, 0);
	if (background_surface) {
		r.x = -((tilemap_x / 4) * scale_w) % (background_surface->w / 2);
		r.y = (screen_h * scale_h - background_surface->h) / 2;
		r.w = background_surface->w;
		r.h = background_surface->h;
		SDL_BlitSurface(background_surface, 0, screen_surface, &r);
	}
	// tiles
	for (int j = 0; j < screen_h + 15; j += TILE_H) {
		for (int i = 0; i < screen_w + 15; i += TILE_W) {
			const uint8_t tile = lookup_tile(i, j);
			if (tile != 0) {
				r.x = scale_w * (i - tilemap_origin_x);
				r.y = scale_h * (j - tilemap_origin_y + screen_voffset);
				r.w = scale_w * TILE_W;
				r.h = scale_h * TILE_H;
				SDL_BlitSurface(sprites_sheets_table[0].surface, &sprites_sheets_table[0].r[tile], screen_surface, &r);
			}
		}
	}
	// sprites
	for (int i = 0; i < sprite_list_size; ++i) {
		const sprite_t *spr = &sprite_list[i];
		spritesheet_t *sheet = &sprites_sheets_table[spr->sheet];
		r.x = scale_w * spr->x;
		r.y = scale_h * spr->y;
		r.w = sheet->r[spr->num].w;
		r.h = sheet->r[spr->num].h;
		SDL_BlitSurface(sheet->surface, &sheet->r[spr->num], screen_surface, &r);
	}
	// fill rects
	for (int i = 0; i < fill_rects_count; ++i) {
		const uint32_t color = color_rgba(screen_surface->format, &fill_rects_table[i].color);
		r = fill_rects_table[i].r;
		r.x *= scale_w;
		r.y *= scale_h;
		r.w *= scale_w;
		r.h *= scale_h;
		SDL_FillRect(screen_surface, &r, color);
	}
}

void sys_update_screen() {
	if (current_mode == SSM_TILED) {
		update_screen_tiled();
	} else {
		SDL_Rect r;
		r.w = background_surface->w;
		r.h = background_surface->h;
		r.x = (screen_w * scale_w - r.w) / 2;
		r.y = (screen_h * scale_h - r.h) / 2;
		SDL_BlitSurface(background_surface, 0, screen_surface, &r);
	}
	// fade
	if (cur_alpha != dst_alpha) {
		if (cur_alpha < dst_alpha) {
			cur_alpha += ALPHA_STEP;
			if (cur_alpha > 255) {
				cur_alpha = 255;
			}
		} else {
			cur_alpha -= ALPHA_STEP;
			if (cur_alpha < 0) {
				cur_alpha = 0;
			}
		}
		SDL_SetAlpha(fade_surface, SDL_SRCALPHA, 255 - cur_alpha);
		SDL_BlitSurface(fade_surface, 0, screen_surface, 0);
	}
	// screen
	SDL_Rect r;
	r.x = 0;
	r.y = 0;
	r.w = screen_surface->w;
	r.h = screen_surface->h;
	SDL_UpdateRect(screen_surface, r.x, r.y, r.w, r.h);
}

void sys_set_screen_offset(int offset) {
	screen_voffset = offset;
}

void sys_fill_rect(const sys_rect_t *rect, int r, int g, int b) {
	if (fill_rects_count < MAX_FILL_RECTS) {
		fill_rect_t *fr = &fill_rects_table[fill_rects_count];
		fr->r.x = rect->x;
		fr->r.y = rect->y;
		fr->r.w = rect->w;
		fr->r.h = rect->h;
		fr->color.a = 255;
		fr->color.r = r;
		fr->color.g = g;
		fr->color.b = b;
		++fill_rects_count;
	} else {
		warning("sys_fill_rect() fill_rects_table overflow");
	}
}

static SDL_Surface *load_graphics(const uint8_t *data, int size, bool transparent) {
	SDL_RWops *rw = SDL_RWFromConstMem(data, size);
	SDL_Surface *surface = IMG_Load_RW(rw, 1);
	if (!surface) {
		warning("Failed to load %d bytes, error '%s'", size, IMG_GetError());
	} else {
		if (transparent) {
			SDL_SetColorKey(surface, SDL_SRCCOLORKEY, SDL_MapRGB(surface->format, 255, 0, 255));
		}
	}
	return surface;
}

void sys_set_tile_data(const uint8_t *src, int size, uint8_t transparent_color) {
	spritesheet_t *sheet = &sprites_sheets_table[0];
	sheet->surface = load_graphics(src, size, true);
	sheet->count = MAX_TILES;
	for (int i = 0; i < MAX_TILES; ++i) {
		SDL_Rect *r = &sheet->r[i];
		r->x = scale_w * i * TILE_W;
		r->y = 0;
		r->w = scale_w * TILE_W;
		r->h = scale_h * TILE_H;
	}
	current_mode = SSM_TILED;
}

void sys_set_tilemap_data(const uint8_t *data, int w, int h, const uint8_t *anim_lut) {
	tilemap_data = data;
	tilemap_lut = anim_lut;
	tilemap_origin_x = 0;
	tilemap_origin_y = 0;
	tilemap_x = 0;
	tilemap_y = 0;
	tilemap_w = w;
	tilemap_h = h;
}

void sys_set_tilemap_origin(int x, int y) {
	tilemap_x = x;
	tilemap_y = y;
	tilemap_origin_x = x & 15;
	tilemap_origin_y = y & 15;
}

void sys_set_copper_pal(const uint16_t *pal_data, int num_lines) {
	SDL_LockSurface(copper.surface);
	draw_copper(copper.surface, pal_data, num_lines);
	SDL_UnlockSurface(copper.surface);
}

void sys_set_background_frame(const uint8_t *frame_data, int size) {
	if (background_surface) {
		SDL_FreeSurface(background_surface);
		background_surface = 0;
	}
	if (frame_data != 0) {
		background_surface = load_graphics(frame_data, size, true);
	}
}

void sys_clear_sprite_data() {
	for (int i = 0; i < sprites_sheets_count; ++i) {
		SDL_FreeSurface(sprites_sheets_table[i].surface);
		memset(&sprites_sheets_table[i], 0, sizeof(spritesheet_t));
	}
	sprites_sheets_count = 0;
}

int sys_load_sprite_data(const uint8_t *data, int size, int count, sys_rect_t *r) {
	++sprites_sheets_count;
	assert(sprites_sheets_count < MAX_SPRITES_SHEETS);
	spritesheet_t *sheet = &sprites_sheets_table[sprites_sheets_count];
	if (count != 0) {
		memset(sheet, 0, sizeof(spritesheet_t));
		sheet->count = count;
		for (int i = 0; i < count; ++i) {
			sheet->r[i].x = scale_w * r[i].x;
			sheet->r[i].y = scale_h * r[i].y;
			sheet->r[i].w = scale_w * r[i].w;
			sheet->r[i].h = scale_h * r[i].h;
		}
	}
	sheet->surface = load_graphics(data, size, true);
	return sprites_sheets_count;
}

void sys_get_sprite_dim(int sheet, int num, int *w, int *h) {
	SDL_Rect *r = &sprites_sheets_table[sheet].r[num];
	*w = r->w / scale_w;
	*h = r->h / scale_h;
}

void sys_clear_sprite_list() {
	sprite_list_size = 0;
	fill_rects_count = 0;
}

void sys_add_to_sprite_list(int spr_data, int num, int x, int y, int flags, int pal_num) {
	if (sprite_list_size < MAX_SPRITES) {
		sprite_t *spr = &sprite_list[sprite_list_size];
		spr->sheet = spr_data;
		spr->num = num;
		spr->x = x;
		spr->y = y;
		spr->flags = flags;
		++sprite_list_size;
	} else {
		warning("sys_add_to_sprite_list() sprite_list overflow");
	}
}

static void sys_process_event(SDL_Event *ev) {
	switch (ev->type) {
	case SDL_QUIT:
		sys_exit_flag = 1;
		break;
	case SDL_ACTIVEEVENT:
		if (ev->active.state & SDL_APPINPUTFOCUS) {
			pause_flag = !ev->active.gain;
		}
		break;
	case SDL_KEYUP:
		debug(DBG_SYSTEM, "SDL_KEYUP %d", ev->key.keysym.sym);
		switch (ev->key.keysym.sym) {
		case SDLK_LEFT:
			pressed_key_mask &= ~SKM_LEFT;
			released_key_mask |= SKM_LEFT;
			break;
		case SDLK_RIGHT:
			pressed_key_mask &= ~SKM_RIGHT;
			released_key_mask |= SKM_RIGHT;
			break;
		case SDLK_UP:
			pressed_key_mask &= ~SKM_UP;
			released_key_mask |= SKM_UP;
			break;
		case SDLK_DOWN:
			pressed_key_mask &= ~SKM_DOWN;
			released_key_mask |= SKM_DOWN;
			break;
		case SDLK_i:
			pressed_key_mask &= ~SKM_INVENTORY;
			released_key_mask |= SKM_INVENTORY;
			break;
		case SDLK_p:
			pressed_key_mask &= ~SKM_PAUSE;
			released_key_mask |= SKM_PAUSE;
			break;
		case SDLK_RCTRL:
		case SDLK_SPACE:
			pressed_key_mask &= ~SKM_ACTION;
			released_key_mask |= SKM_ACTION;
			break;
		case SDLK_F1:
			pressed_key_mask &= ~SKM_F1;
			released_key_mask |= SKM_F1;
			break;
		case SDLK_F2:
			pressed_key_mask &= ~SKM_F2;
			released_key_mask |= SKM_F2;
			break;
		case SDLK_F3:
			pressed_key_mask &= ~SKM_F3;
			released_key_mask |= SKM_F3;
			break;
		case SDLK_F4:
			pressed_key_mask &= ~SKM_F4;
			released_key_mask |= SKM_F4;
			break;
		case SDLK_F5:
			pressed_key_mask &= ~SKM_F5;
			released_key_mask |= SKM_F5;
			break;
		case SDLK_F6:
			pressed_key_mask &= ~SKM_F6;
			released_key_mask |= SKM_F6;
			break;
		case SDLK_F7:
			pressed_key_mask &= ~SKM_F7;
			released_key_mask |= SKM_F7;
			break;
		case SDLK_F8:
			pressed_key_mask &= ~SKM_F8;
			released_key_mask |= SKM_F8;
			break;
		case SDLK_ESCAPE:
			pressed_key_mask &= ~SKM_QUIT;
			released_key_mask |= SKM_QUIT;
			break;
		default:
			break;
		}
		break;
	case SDL_KEYDOWN:
		debug(DBG_SYSTEM, "SDL_KEYDOWN %d", ev->key.keysym.sym);
		if (ev->key.keysym.mod & KMOD_ALT) {
			switch (ev->key.keysym.sym) {
			case SDLK_RETURN:
				change_screen_surface(!fullscreen_flag);
				break;
			default:
				break;
			}
			break;
		}
		switch (ev->key.keysym.sym) {
		case SDLK_LEFT:
			pressed_key_mask |= SKM_LEFT;
			break;
		case SDLK_RIGHT:
			pressed_key_mask |= SKM_RIGHT;
			break;
		case SDLK_UP:
			pressed_key_mask |= SKM_UP;
			break;
		case SDLK_DOWN:
			pressed_key_mask |= SKM_DOWN;
			break;
		case SDLK_i:
			pressed_key_mask |= SKM_INVENTORY;
			break;
		case SDLK_p:
			pressed_key_mask |= SKM_PAUSE;
			break;
		case SDLK_RCTRL:
		case SDLK_SPACE:
			pressed_key_mask |= SKM_ACTION;
			break;
		case SDLK_F1:
			pressed_key_mask |= SKM_F1;
			break;
		case SDLK_F2:
			pressed_key_mask |= SKM_F2;
			break;
		case SDLK_F3:
			pressed_key_mask |= SKM_F3;
			break;
		case SDLK_F4:
			pressed_key_mask |= SKM_F4;
			break;
		case SDLK_F5:
			pressed_key_mask |= SKM_F5;
			break;
		case SDLK_F6:
			pressed_key_mask |= SKM_F6;
			break;
		case SDLK_F7:
			pressed_key_mask |= SKM_F7;
			break;
		case SDLK_F8:
			pressed_key_mask |= SKM_F8;
			break;
		case SDLK_ESCAPE:
			pressed_key_mask |= SKM_QUIT;
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
}

void sys_process_events() {
	int ms;
	SDL_Event ev;
	released_key_mask = 0;
	while (sys_exit_flag == 0) {
		while (SDL_PollEvent(&ev)) {
			sys_process_event(&ev);
		}
		if (!sys_exit_flag && !pause_flag) {
			ms = (SDL_GetTicks() - last_frame_time_stamp) * TICK_DIV;
			if (ms >= TICK_MUL) {
				last_frame_time_stamp = SDL_GetTicks();
				break;
			}
		}
		SDL_Delay(10);
	}
}

int sys_get_keys_pressed_mask() {
	return pressed_key_mask;
}

int sys_get_keys_released_mask() {
	return released_key_mask;
}

void sys_play_sfx(int num, const uint8_t *data, int size) {
	SDL_RWops *rw = SDL_RWFromConstMem(data, size);
	sound = Mix_LoadWAV_RW(rw, 1);
	if (sound) {
		Mix_PlayChannel(-1, sound, 0);
	}
}

void sys_play_module(int num, const uint8_t *data, int size) {
#ifndef __EMSCRIPTEN__
	SDL_RWops *rw = SDL_RWFromConstMem(data, size);
	music = Mix_LoadMUSType_RW(rw, MUS_MOD, 1);
	if (music) {
		Mix_VolumeMusic(MIX_MAX_VOLUME / 2);
		Mix_PlayMusic(music, 0);
	}
#endif
}

void sys_stop_module() {
	if (music) {
		Mix_HaltMusic();
		Mix_FreeMusic(music);
		music = 0;
	}
}

void sys_print_string(const char *str) {
	fprintf(stdout, "%s", str);
}
